#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <pthread.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include "IOTCAPIs.h"
#include "AVAPIs.h"
#include "P2PCam/AVFRAMEINFO.h"
#include "P2PCam/AVIOCTRLDEFs.h"
#include "token/token.h"
#include "demoOption/demoOption.h"
#include "sample_constants.h"

// *** set to 1 to enable audio or speaker function ***
#define ENABLE_AUDIO    1
#define ENABLE_SPEAKER  1
#define ENABLE_RESEND   1

#define AUDIO_BUF_SIZE	1024
#define VIDEO_BUF_SIZE	128000

#define CHANGE_PASSWORD_TIMEOUT_SEC 100
#define DELETE_IDENTITY_TIMEOUT_SEC 100
#define LIST_IDENTITIES_TIMEOUT_SEC 100
#define REQUEST_TOKEN_TIMEOUT_SEC   100
#define RECV_IO_CONTROL_TIMEOUT_SEC 100
#define REQUEST_ABILITY_TIMEOUT_SEC 100
#define SEARCH_DEVICE_TIME 6000
#define SEARCH_DEVICE_TIME_ONE_ROUND 100

    /*
 * Default DemoConfig :
 *      iotc_auth_key is "00000000"
 *      av_account    is "admin"
 *      av_password   is "888888"
 *      av_identity   is "admin"
 *      av_token      is "888888"
 */
DemoConfig gDemoConfig;
DemoOptions gDemoOptions;

struct timeval gConnectStartTv;

typedef struct _AV_Server
{
    char UID[21];
    int SID;
    int speakerCh;
    int twoWayStream;
    int avIndex;
}AV_Server;

void PrintErrHandling (int nErr)
{
	switch (nErr)
	{
	case IOTC_ER_SERVER_NOT_RESPONSE :
		//-1 IOTC_ER_SERVER_NOT_RESPONSE
		printf ("[Error code : %d]\n", IOTC_ER_SERVER_NOT_RESPONSE );
		printf ("Master doesn't respond.\n");
		printf ("Please check the network wheather it could connect to the Internet.\n");
		break;
	case IOTC_ER_FAIL_RESOLVE_HOSTNAME :
		//-2 IOTC_ER_FAIL_RESOLVE_HOSTNAME
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_RESOLVE_HOSTNAME);
		printf ("Can't resolve hostname.\n");
		break;
	case IOTC_ER_ALREADY_INITIALIZED :
		//-3 IOTC_ER_ALREADY_INITIALIZED
		printf ("[Error code : %d]\n", IOTC_ER_ALREADY_INITIALIZED);
		printf ("Already initialized.\n");
		break;
	case IOTC_ER_FAIL_CREATE_MUTEX :
		//-4 IOTC_ER_FAIL_CREATE_MUTEX
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_MUTEX);
		printf ("Can't create mutex.\n");
		break;
	case IOTC_ER_FAIL_CREATE_THREAD :
		//-5 IOTC_ER_FAIL_CREATE_THREAD
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_THREAD);
		printf ("Can't create thread.\n");
		break;
	case IOTC_ER_FAIL_CREATE_SOCKET :
		//-6 IOTC_ER_FAIL_CREATE_SOCKET
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_SOCKET);
		printf ("Fails to create sockets.\n");
		break;
	case IOTC_ER_FAIL_SOCKET_OPT :
		//-7 IOTC_ER_FAIL_SOCKET_OPT
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_SOCKET_OPT);
		printf ("Fails to set up socket options.\n");
		break;
	case IOTC_ER_FAIL_SOCKET_BIND :
		//-8 IOTC_ER_FAIL_SOCKET_BIND
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_SOCKET_BIND);
		printf ("Fails to bind sockets.\n");
		break;
	case IOTC_ER_UNLICENSE :
		//-10 IOTC_ER_UNLICENSE
		printf ("[Error code : %d]\n", IOTC_ER_UNLICENSE);
		printf ("This UID is unlicense.\n");
		printf ("Check your UID.\n");
		break;
	case IOTC_ER_NOT_INITIALIZED :
		//-12 IOTC_ER_NOT_INITIALIZED
		printf ("[Error code : %d]\n", IOTC_ER_NOT_INITIALIZED);
		printf ("Please initialize the IOTCAPI first.\n");
		break;
	case IOTC_ER_TIMEOUT :
		//-13 IOTC_ER_TIMEOUT
		break;
	case IOTC_ER_INVALID_SID :
		//-14 IOTC_ER_INVALID_SID
		printf ("[Error code : %d]\n", IOTC_ER_INVALID_SID);
		printf ("This SID is invalid.\n");
		printf ("Please check it again.\n");
		break;
	case IOTC_ER_FAIL_GET_LOCAL_IP :
		//-16 IOTC_ER_FAIL_GET_LOCAL_IP
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_GET_LOCAL_IP);
		printf ("Fails to get the client's local IP address.\n");
		break;
	case IOTC_ER_EXCEED_MAX_SESSION :
		//-18 IOTC_ER_EXCEED_MAX_SESSION
		printf ("[Error code : %d]\n", IOTC_ER_EXCEED_MAX_SESSION);
		printf ("[Warning]\n");
		printf ("The amount of session reach to the maximum.\n");
		printf ("It cannot be connected unless the session is released.\n");
		break;
	case IOTC_ER_CAN_NOT_FIND_DEVICE :
		//-19 IOTC_ER_CAN_NOT_FIND_DEVICE
		printf ("[Error code : %d]\n", IOTC_ER_CAN_NOT_FIND_DEVICE);
		printf ("Device didn't register on server, so we can't find device.\n");
		printf ("Please check the device again.\n");
		printf ("Retry...\n");
		break;
	case IOTC_ER_SESSION_CLOSE_BY_REMOTE :
		//-22 IOTC_ER_SESSION_CLOSE_BY_REMOTE
		printf ("[Error code : %d]\n", IOTC_ER_SESSION_CLOSE_BY_REMOTE);
		printf ("Session is closed by remote so we can't access.\n");
		printf ("Please close it or establish session again.\n");
		break;
	case IOTC_ER_REMOTE_TIMEOUT_DISCONNECT :
		//-23 IOTC_ER_REMOTE_TIMEOUT_DISCONNECT
		printf ("[Error code : %d]\n", IOTC_ER_REMOTE_TIMEOUT_DISCONNECT);
		printf ("We can't receive an acknowledgement character within a TIMEOUT.\n");
		printf ("It might that the session is disconnected by remote.\n");
		printf ("Please check the network wheather it is busy or not.\n");
		printf ("And check the device and user equipment work well.\n");
		break;
	case IOTC_ER_DEVICE_NOT_LISTENING :
		//-24 IOTC_ER_DEVICE_NOT_LISTENING
		printf ("[Error code : %d]\n", IOTC_ER_DEVICE_NOT_LISTENING);
		printf ("Device doesn't listen or the sessions of device reach to maximum.\n");
		printf ("Please release the session and check the device wheather it listen or not.\n");
		break;
	case IOTC_ER_CH_NOT_ON :
		//-26 IOTC_ER_CH_NOT_ON
		printf ("[Error code : %d]\n", IOTC_ER_CH_NOT_ON);
		printf ("Channel isn't on.\n");
		printf ("Please open it by IOTC_Session_Channel_ON() or IOTC_Session_Get_Free_Channel()\n");
		printf ("Retry...\n");
		break;
	case IOTC_ER_FAIL_CONNECT_SEARCH:
		//-27 IOTC_ER_FAIL_CONNECT_SEARCH
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_CONNECT_SEARCH);
		printf ("The client stop connecting to the device.\n");
		break;
	case IOTC_ER_SESSION_NO_FREE_CHANNEL :
		//-31 IOTC_ER_SESSION_NO_FREE_CHANNEL
		printf ("[Error code : %d]\n", IOTC_ER_SESSION_NO_FREE_CHANNEL);
		printf ("All channels are occupied.\n");
		printf ("Please release some channel.\n");
		break;
	case IOTC_ER_TCP_TRAVEL_FAILED :
		//-32 IOTC_ER_TCP_TRAVEL_FAILED
		printf ("[Error code : %d]\n", IOTC_ER_TCP_TRAVEL_FAILED);
		printf ("Device can't connect to Master.\n");
		printf ("Don't let device use proxy.\n");
		printf ("Close firewall of device.\n");
		printf ("Or open device's TCP port 80, 443, 8080, 8000, 21047.\n");
		break;
	case IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED :
		//-33 IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED
		printf ("[Error code : %d]\n", IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED);
		printf ("Device can't connect to server by TCP.\n");
		printf ("Don't let server use proxy.\n");
		printf ("Close firewall of server.\n");
		printf ("Or open server's TCP port 80, 443, 8080, 8000, 21047.\n");
		printf ("Retry...\n");
		break;
	case IOTC_ER_NO_PERMISSION :
		//-40 IOTC_ER_NO_PERMISSION
		printf ("[Error code : %d]\n", IOTC_ER_NO_PERMISSION);
		printf ("This UID's license doesn't support TCP.\n");
		break;
	case IOTC_ER_NETWORK_UNREACHABLE :
		//-41 IOTC_ER_NETWORK_UNREACHABLE
		printf ("[Error code : %d]\n", IOTC_ER_NETWORK_UNREACHABLE);
		printf ("Network is unreachable.\n");
		printf ("Please check your network.\n");
		printf ("Retry...\n");
		break;
	case IOTC_ER_FAIL_SETUP_RELAY :
		//-42 IOTC_ER_FAIL_SETUP_RELAY
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_SETUP_RELAY);
		printf ("Client can't connect to a device via Lan, P2P, and Relay mode\n");
		break;
	case IOTC_ER_NOT_SUPPORT_RELAY :
		//-43 IOTC_ER_NOT_SUPPORT_RELAY
		printf ("[Error code : %d]\n", IOTC_ER_NOT_SUPPORT_RELAY);
		printf ("Server doesn't support UDP relay mode.\n");
		printf ("So client can't use UDP relay to connect to a device.\n");
		break;
	case IOTC_ER_NO_SERVER_LIST :
		//-44 IOTC_ER_NO_SERVER_LIST
		printf ("[Error code : %d]\n", IOTC_ER_NO_SERVER_LIST);
		printf ("No IOTC server information while client connect.\n");
		break;
	case IOTC_ER_DEVICE_MULTI_LOGIN :
		//-45 IOTC_ER_DEVICE_MULTI_LOGIN
		printf ("[Error code : %d]\n", IOTC_ER_DEVICE_MULTI_LOGIN);
		printf ("The connecting device duplicated loggin and may unconnectable.\n");
		break;
	case IOTC_ER_MASTER_NOT_RESPONSE :
		//-60 IOTC_ER_MASTER_NOT_RESPONSE
		printf ("[Error code : %d]\n", IOTC_ER_MASTER_NOT_RESPONSE);
		printf ("All masters have no respond.\n");
		break;
	case IOTC_ER_DEVICE_IS_SLEEP :
		//-64 IOTC_ER_DEVICE_IS_SLEEP
		printf ("[Error code : %d]\n", IOTC_ER_DEVICE_IS_SLEEP);
		printf ("Device is in sleep mode.\n");
		break;
	case IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY :
		// -68 IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY
		printf ("[Error code : %d]\n", IOTC_ER_DEVICE_REJECT_BY_WRONG_AUTH_KEY);
		printf ("Being rejected by device due to the wrong auth key.\n");
		break;
	case IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION :
		// -69 IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION
		printf ("[Error code : %d]\n", IOTC_ER_DEVICE_NOT_USE_KEY_AUTHENTICATION);
		printf ("Being rejected by device due to disable authentication.\n");
		break;

	default :
		printf("[Error code : %d]\n", nErr);
		break;
	}
}

int ExPasswordAuthCallBackFn(const char *account, char *pwd, unsigned int pwd_buf_size)
{
    if (strncmp(account, gDemoConfig.av_account, strlen(gDemoConfig.av_account)) != 0)
        return -1;

    if (pwd_buf_size <= strlen(gDemoConfig.av_password)) return -1;
    strcpy(pwd, gDemoConfig.av_password);

    return 0;
}

int ExTokenAuthCallBackFn(const char *identity, char *token, unsigned int token_buf_size)
{
    if(strncmp(identity, gDemoConfig.av_identity, strlen(gDemoConfig.av_identity)) != 0)
        return -1;

    if (token_buf_size <= strlen(gDemoConfig.av_token)) return -1;
    strcpy(token, gDemoConfig.av_token);

    return 0;
}

void *thread_Speaker(void *arg)
{
	AV_Server *ServerInfo = (AV_Server *)arg;
	char buf[AUDIO_BUF_SIZE];
	int frameRate = 25;
	int sleepTick = 1000000/frameRate;
	//int resend = 0;
    int avIndex = 0;
	//int sendCnt = 0;
	FILE *fp = fopen("./audio/8k_16bit_mono.raw", "rb");
	if(fp == NULL)
	{
		printf("fopen error!!!\n");
		return 0;
	}

	FRAMEINFO_t frameInfo;
	memset(&frameInfo, 0, sizeof(frameInfo));

    if(1 == ServerInfo->twoWayStream)
    {
        avIndex = ServerInfo->avIndex;
    } else {
        AVServStartInConfig avStartInConfig;
        AVServStartOutConfig avStartOutConfig;

        memset(&avStartInConfig, 0, sizeof(avStartInConfig));
        avStartInConfig.cb               = sizeof(AVServStartInConfig);
        avStartInConfig.iotc_session_id  = ServerInfo->SID;
        avStartInConfig.iotc_channel_id  = ServerInfo->speakerCh;
        avStartInConfig.timeout_sec      = 30;
        avStartInConfig.password_auth    = &ExPasswordAuthCallBackFn;
        avStartInConfig.token_auth       = &ExTokenAuthCallBackFn;
        avStartInConfig.server_type      = 0;
        avStartInConfig.resend           = ENABLE_RESEND;
        if (gDemoOptions.enable_dtls)
            avStartInConfig.security_mode = AV_SECURITY_DTLS;
        else
            avStartInConfig.security_mode = AV_SECURITY_SIMPLE;

        avStartOutConfig.cb              = sizeof(AVServStartOutConfig);

        avIndex = avServStartEx(&avStartInConfig, &avStartOutConfig);
        if(avIndex < 0)
        {
            printf("avServStartEx failed[%d]\n", avIndex);
            return 0;
        }
    }

	printf("[thread_Speaker] Starting avIndex[%d] \n", avIndex);

	frameInfo.codec_id = MEDIA_CODEC_AUDIO_PCM;
	frameInfo.flags = (AUDIO_SAMPLE_8K << 2) | (AUDIO_DATABITS_16 << 1) | AUDIO_CHANNEL_MONO;

		while(1)
		{
			int size = fread(buf, 1, 640, fp);
			if(size <= 0)
			{
				printf("rewind audio\n");
				rewind(fp);
				continue;
			}

			int ret = avSendAudioData(avIndex, buf, 640, &frameInfo, sizeof(FRAMEINFO_t));
			if(ret == AV_ER_SESSION_CLOSE_BY_REMOTE)
			{
				printf("thread_AudioFrameData AV_ER_SESSION_CLOSE_BY_REMOTE\n");
				break;
			}
			else if(ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT)
			{
				printf("thread_AudioFrameData AV_ER_REMOTE_TIMEOUT_DISCONNECT\n");
				break;
			}
			else if(ret == IOTC_ER_INVALID_SID)
			{
				printf("Session cant be used anymore\n");
				break;
			}
			else if(ret < 0)
			{
				printf("avSendAudioData error[%d]\n", ret);
				break;
			}
			usleep(sleepTick);

			//*** Speaker thread stop condition if necessary ***
/*
			if(sendCnt++ > 1500)
				break;
*/
		}

	fclose(fp);
	printf("[thread_Speaker] exit...\n");
	if(1 != ServerInfo->twoWayStream)
        avServStop(avIndex);

	return 0;
}


void *thread_ReceiveAVData(void *arg)
{
	printf("[%s] Starting....\n", __func__);
	int avIndex = *(int *)arg;
    int video_buf_size = VIDEO_BUF_SIZE;
	char *buf = (char*)malloc(video_buf_size);
	int ret;

	FRAMEINFO_t frameInfo;
	char *video_format[] = {"MPEG4", "H263", "H264", "MJPEG", "UNKNOWN"};
    char *audio_format[] = {"ADPCM", "PCM", "SPEEX", "MP3", "G726", "UNKNOWN"};
	int codec_idx = 0;
	unsigned int frmNo;
	struct timeval tv, tv2;

	int vFPS = 0, aFPS = 0, round = 0, vLostCnt = 0, aLostCnt = 0;
	int outBufSize = 0;
	int outFrmSize = 0;
	int outFrmInfoSize = 0;
	int vbps = 0, abps = 0;
    int video_not_ready = 1, audio_not_ready = 1;
	gettimeofday(&tv, NULL);

	while(1)
	{
		ret = avRecvFrameData2(avIndex, buf, video_buf_size, &outBufSize, &outFrmSize, (char *)&frameInfo, sizeof(FRAMEINFO_t), &outFrmInfoSize, &frmNo);
		// show Frame Info at 1st frame
		if(ret > 0 && frmNo == 0) {
			if(frameInfo.codec_id == MEDIA_CODEC_VIDEO_MPEG4)
				codec_idx = 0;
			else if(frameInfo.codec_id == MEDIA_CODEC_VIDEO_H263)
				codec_idx = 1;
			else if(frameInfo.codec_id == MEDIA_CODEC_VIDEO_H264)
				codec_idx = 2;
			else if(frameInfo.codec_id == MEDIA_CODEC_VIDEO_MJPEG)
				codec_idx = 3;
			else
				codec_idx = 4;

			printf("--- Recv Video Format: %s ---\n", video_format[codec_idx]);
		}

		if(ret == AV_ER_DATA_NOREADY) {
			//printf("Recv video, AV_ER_DATA_NOREADY[%d]\n", avIndex);
			//usleep(5 * 1000);
            video_not_ready = 1;
		} else if(ret == AV_ER_LOSED_THIS_FRAME) {
			printf("Recv video, Lost video frame NO[%d]\n", frmNo);
			vLostCnt++;
		} else if(ret == AV_ER_INCOMPLETE_FRAME) {
			#if 1
			if(outFrmInfoSize > 0)
			printf("Incomplete video frame NO[%d] ReadSize[%d] FrmSize[%d] FrmInfoSize[%u] Codec[%d] Flag[%d]\n", frmNo, outBufSize, outFrmSize, outFrmInfoSize, frameInfo.codec_id, frameInfo.flags);
			else
			printf("Incomplete video frame NO[%d] ReadSize[%d] FrmSize[%d] FrmInfoSize[%u]\n", frmNo, outBufSize, outFrmSize, outFrmInfoSize);
			#endif
			vLostCnt++;
		} else if(ret == AV_ER_SESSION_CLOSE_BY_REMOTE) {
			printf("Recv video, AV_ER_SESSION_CLOSE_BY_REMOTE\n");
			break;
		} else if(ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT) {
			printf("Recv video, AV_ER_REMOTE_TIMEOUT_DISCONNECT\n");
			break;
		} else if(ret == IOTC_ER_INVALID_SID) {
			printf("Recv video, Session cant be used anymore\n");
			break;
		} else if(ret == AV_ER_BUFPARA_MAXSIZE_INSUFF) {
            printf("Recv video, video buffer too small\n");
            video_buf_size = video_buf_size+(5*1024);
            buf = (char*)realloc(buf, video_buf_size);
            printf("increase buffer size to %d\n", video_buf_size);
            continue;
        } else {
			vbps += outBufSize;
            vFPS++;
            video_not_ready = 0;
		}

#if ENABLE_AUDIO
        ret = avRecvAudioData(avIndex, buf, AUDIO_BUF_SIZE, (char *)&frameInfo, sizeof(FRAMEINFO_t), &frmNo);
		// show Frame Info at 1st frame
		if(ret > 0 && frmNo == 0)
		{
			int idx = 0;
			if(frameInfo.codec_id == MEDIA_CODEC_AUDIO_ADPCM)
				idx = 0;
			else if(frameInfo.codec_id == MEDIA_CODEC_AUDIO_PCM)
				idx = 1;
			else if(frameInfo.codec_id == MEDIA_CODEC_AUDIO_SPEEX)
				idx = 2;
			else if(frameInfo.codec_id == MEDIA_CODEC_AUDIO_MP3)
				idx = 3;
			else if(frameInfo.codec_id == MEDIA_CODEC_AUDIO_G726)
				idx = 4;
			else
				idx = 5;
			printf("--- Recv Audio Format: %s ---\n", audio_format[idx]);
		}

		if(ret == AV_ER_SESSION_CLOSE_BY_REMOTE) {
			printf("Recv audio, AV_ER_SESSION_CLOSE_BY_REMOTE\n");
			break;
		} else if(ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT) {
			printf("Recv audio, AV_ER_REMOTE_TIMEOUT_DISCONNECT\n");
			break;
		} else if(ret == IOTC_ER_INVALID_SID) {
			printf("Recv audio, Session cant be used anymore\n");
			break;
		} else if(ret == AV_ER_LOSED_THIS_FRAME) {
            printf("Recv audio, AV_ER_LOSED_THIS_FRAME[%d]\n", frmNo);
            aLostCnt++;
		} else if (ret == AV_ER_DATA_NOREADY) {
            //printf("Recv audio, AV_ER_DATA_NOREADY[%d]\n", avIndex);
            //usleep(5 * 1000);
            audio_not_ready = 1;
        } else if(ret < 0) {
			printf("Recv audio, error[%d]!!!\n", ret);
			continue;
		} else {
            abps += ret;
            aFPS++;
            audio_not_ready = 0;
        }
#endif
		gettimeofday(&tv2, NULL);
		long sec = tv2.tv_sec-tv.tv_sec, usec = tv2.tv_usec-tv.tv_usec;
		if(usec < 0) {
			sec--;
			usec += 1000000;
		}
		usec += (sec*1000000);

		if(usec > 1000000) {
			round++;
			printf("[AVAPIs_Client][avIndex:%d] Video FPS=%d, LostFrmCnt:%d, bps:%d Kbps, Audio FPS=%d, LostFrmCnt:%d, bps:%d Kbps\n",
				   avIndex, vFPS, vLostCnt, (vbps/1024)*8, aFPS, aLostCnt, (abps/1024)*8);
			gettimeofday(&tv, NULL);
			vbps = 0;
            abps = 0;
            vFPS = 0;
            aFPS = 0;
		}
        if(1 == video_not_ready && 1 == audio_not_ready)
            usleep(5*1000);
	}

    if(buf) {
        free(buf);
    }
	printf("[%s] thread exit\n", __func__);

	return 0;
}

/*********
Send IOCtrl CMD to device
***/
int start_ipcam_stream(int avIndex, AV_Server *ServerInfo)
{
	int ret;

	SMsgAVIoctrlAVStream ioMsg;
	memset(&ioMsg, 0, sizeof(SMsgAVIoctrlAVStream));
	if((ret = avSendIOCtrl(avIndex, IOTYPE_USER_IPCAM_START, (char *)&ioMsg, sizeof(SMsgAVIoctrlAVStream))) < 0)
	{
		printf("start_ipcam_stream failed[%d]\n", ret);
		return 0;
	}
	printf("send Cmd: IOTYPE_USER_IPCAM_START, OK\n");

#if ENABLE_AUDIO
	if((ret = avSendIOCtrl(avIndex, IOTYPE_USER_IPCAM_AUDIOSTART, (char *)&ioMsg, sizeof(SMsgAVIoctrlAVStream))) < 0)
	{
		printf("start_ipcam_stream failed[%d]\n", ret);
		return 0;
	}
	printf("send Cmd: IOTYPE_USER_IPCAM_AUDIOSTART, OK\n");
#endif

#if ENABLE_SPEAKER
	ioMsg.channel = ServerInfo->speakerCh;
	if((ret = avSendIOCtrl(avIndex, IOTYPE_USER_IPCAM_SPEAKERSTART, (char *)&ioMsg, sizeof(SMsgAVIoctrlAVStream))) < 0)
	{
		printf("start_ipcam_stream failed[%d]\n", ret);
		return 0;
	}
	printf("send Cmd: IOTYPE_USER_IPCAM_SPEAKERSTART, OK\n");
#endif

	return 1;
}

void DemoLanSearch()
{
	struct st_SearchDeviceInfo psSearchDeviceInfo[10];
	IOTC_Search_Device_Start(SEARCH_DEVICE_TIME, SEARCH_DEVICE_TIME_ONE_ROUND);
	while(1)
	{
		int i;
		sleep(1);
		int nRet =IOTC_Search_Device_Result(psSearchDeviceInfo, 10, 0);
		if(nRet < 0)
		{
			break;
		}
		for(i=0;i<nRet;i++)
		{
			printf("UID[%s] Addr[%s:%d]\n", psSearchDeviceInfo[i].UID, psSearchDeviceInfo[i].IP, psSearchDeviceInfo[i].port);
		}
	}
    printf("LAN search done...\n");
}

int DemoChangePassword(int avIndex)
{
    int ret = 0;
    char authkey[IOTC_AUTH_KEY_LENGTH+1] = {0};
    unsigned int length = IOTC_AUTH_KEY_LENGTH;

    ret = avClientRequestChangeServerPassword(avIndex, gDemoConfig.av_account, gDemoConfig.av_password,
            gDemoOptions.new_pass, authkey, sizeof(authkey), &length, CHANGE_PASSWORD_TIMEOUT_SEC);
    if (ret == AV_ER_NoERROR) {
        if (strlen(authkey) > 0) {
            printf("    [%s:%d] change password success and get new authkey[%s] from server\n", __FUNCTION__, __LINE__, authkey);
            strncpy(gDemoConfig.iotc_auth_key, authkey, IOTC_AUTH_KEY_LENGTH+1);
        } else {
            printf("    [%s:%d] change password success without authkey\n", __FUNCTION__, __LINE__);
        }

        //save config
        strncpy(gDemoConfig.av_password, gDemoOptions.new_pass, NEW_MAXSIZE_VIEWPWD);
        saveConfig(&gDemoConfig);
    } else {
        printf("    [%s:%d] change password failed ret[%d]\n", __FUNCTION__, __LINE__, ret);
    }

    return ret;
}

int DemoCreateNewToken(int avIndex)
{
    int status, ret;
    char new_token[MAX_TOKEN_LENGTH+1];
    ret = avClientRequestTokenWithIdentity(avIndex,
            gDemoOptions.new_identity, "guest", new_token,
            sizeof(new_token), &status, REQUEST_TOKEN_TIMEOUT_SEC);

	if(ret < AV_ER_NoERROR || status != TOKEN_SUCCESS) {
		printf("[%s:%d] avClientRequestTokenWithIdentity fail ret[%d] status[%d]\n",  __FUNCTION__, __LINE__, ret, status);
		return ret;
	}

	 // you can define your own status here
    printf("[%s:%d] create new_identity[%s] with token[%s] len[%lu] status[%d]\n", __FUNCTION__, __LINE__,
        gDemoOptions.new_identity, new_token, strlen(new_token), status);

    //save config
    strncpy(gDemoConfig.av_identity, gDemoOptions.new_identity, MAX_IDENTITY_LENGTH+1);
    strncpy(gDemoConfig.av_token, new_token, MAX_TOKEN_LENGTH+1);
    saveConfig(&gDemoConfig);

    return ret;
}

int DemoDeleteIdentity(int avIndex)
{
    int status, ret;
    ret = avClientRequestDeleteTokenWithIdentity(avIndex, gDemoOptions.del_identity,
            &status, DELETE_IDENTITY_TIMEOUT_SEC);
    if(ret == 0)
        printf("    [%s:%d] delete identity[%s] success with status[%d]\n", __FUNCTION__, __LINE__,
            gDemoOptions.del_identity, status);
    else
        printf("    [%s:%d] delete identity[%s] failed ret[%d]\n", __FUNCTION__, __LINE__,
            gDemoOptions.del_identity, ret);

    return ret;
}

int DemoListIdentity(int avIndex)
{
    int status, ret, i;
    AvIdentity *identities;
    unsigned int identity_count = 0;
    ret = avClientRequestIdentityArray(avIndex, &identities,
            &identity_count, &status, LIST_IDENTITIES_TIMEOUT_SEC);
    if(ret == 0){
        printf("    [%s:%d] get identities with status[%d] count[%d]\n", __FUNCTION__, __LINE__, status, identity_count);
        for(i = 0 ; i < identity_count ; i++)
            printf("        [%d] identity [%s]\n", i, identities[i].identity);
        avClientFreeIdentityArray(identities);
        identities = NULL;
    }
    else
        printf("    [%s:%d] delete identity failed ret[%d]\n", __FUNCTION__, __LINE__, ret);

    return ret;
}

int printServerAbility(int avIndex)
{
    int ret;
    unsigned char *ability = NULL;
    unsigned int ability_length = 0;
    ret = avClientRequestServerAbility(avIndex, &ability, &ability_length, REQUEST_ABILITY_TIMEOUT_SEC);
    if (ret == 0) {
        printf("    [%s:%d] get server ability\n", __FUNCTION__, __LINE__);
        printf("====================\n%s\n====================\n", (char*)ability);
        avClientFreeServerAbility(ability);
        ability = NULL;
    }
    else
        printf("    [%s:%d] get ability failed ret[%d]\n", __FUNCTION__, __LINE__, ret);

    return ret;
}

void *thread_ConnectCCR(void *arg)
{
    int ret, SID=-1, tmpSID, avIndex=-1;
    pthread_t ThreadAVData_ID = 0, ThreadSpeaker_ID = 0;
    AV_Server *ServerInfo = (AV_Server *)arg;

    tmpSID = IOTC_Get_SessionID();
    if(tmpSID < 0) {
        printf("  [%s:%d] IOTC_Get_SessionID failed ret[%d]\n", __FUNCTION__, __LINE__, tmpSID);
        goto __thread_ConnectCCR_exit;
    }
    printf("  [%s:%d] IOTC_Get_SessionID tmpSID[%d]\n", __FUNCTION__, __LINE__, tmpSID);

    IOTCConnectInput connect_input = {0};
    connect_input.cb = sizeof(connect_input);
    connect_input.authentication_type = AUTHENTICATE_BY_KEY;
    connect_input.timeout = 20; //sec
    memcpy(connect_input.auth_key, gDemoConfig.iotc_auth_key, IOTC_AUTH_KEY_LENGTH);

    SID = IOTC_Connect_ByUIDEx(ServerInfo->UID, tmpSID, &connect_input);

    if (SID < 0) {
        printf("  [%s:%d] %s failed SID[%d]\n", __FUNCTION__, __LINE__, "IOTC_Connect_ByUIDEx", SID);
        PrintErrHandling (SID);
        goto __thread_ConnectCCR_exit;
    }

	struct timeval tv2;
	gettimeofday(&tv2, NULL);
	unsigned long msec = (tv2.tv_sec-gConnectStartTv.tv_sec)*1000+(tv2.tv_usec-gConnectStartTv.tv_usec)/1000;
	printf("  [%s:%d] %s to %s SID[%d] Cost Time[%lu]ms\n", __FUNCTION__, __LINE__,
		"IOTC_Connect_ByUIDEx", ServerInfo->UID, SID, msec);

	struct st_SInfoEx Sinfo;
	memset(&Sinfo, 0, sizeof(struct st_SInfoEx));
	Sinfo.size = sizeof(struct st_SInfoEx);
	ret = IOTC_Session_Check_Ex(SID, &Sinfo);
	if(ret != IOTC_ER_NoERROR) {
		printf("IOTC_Session_Check_Ex occur ERROR %d on %d %s\n",ret, __LINE__, __FUNCTION__);
		goto __thread_ConnectCCR_exit;
	}

	char *mode[] = {"P2P", "RLY", "LAN"};
	ret = isdigit( Sinfo.RemoteIP[0]);
	if(ret == 0){
		printf("isdigit() occur ERROR %d on %d %s\n",ret, __LINE__, __FUNCTION__);
	}

	printf("  [%s:%d] Device[%s] is from [%s:%d] Mode[%s] NatType[%d] IOTCVersion[%X]\n", __FUNCTION__, __LINE__,
		Sinfo.UID, Sinfo.RemoteIP, Sinfo.RemotePort, mode[(int)Sinfo.Mode], Sinfo.RemoteNatType, Sinfo.IOTCVersion);

    AVClientStartInConfig avClientInConfig;
    AVClientStartOutConfig avClientOutConfig;
    memset(&avClientInConfig, 0, sizeof(avClientInConfig));
    avClientInConfig.cb               = sizeof(AVClientStartInConfig);
    avClientInConfig.iotc_channel_id  = 0;
    avClientInConfig.iotc_session_id  = SID;
    avClientInConfig.timeout_sec      = 30;
    avClientInConfig.resend           = ENABLE_RESEND;
    if (gDemoOptions.enable_dtls)
        avClientInConfig.security_mode = AV_SECURITY_DTLS;
    else
        avClientInConfig.security_mode = AV_SECURITY_SIMPLE;

    if (gDemoOptions.enable_connect_with_token) {
        avClientInConfig.account_or_identity    = gDemoConfig.av_identity;
        avClientInConfig.password_or_token      = gDemoConfig.av_token;
        avClientInConfig.auth_type              = AV_AUTH_TOKEN;
    } else {
        avClientInConfig.account_or_identity    = gDemoConfig.av_account;
        avClientInConfig.password_or_token      = gDemoConfig.av_password;
        avClientInConfig.auth_type              = AV_AUTH_PASSWORD;
    }
    avClientOutConfig.cb = sizeof(AVClientStartOutConfig);

    avIndex = avClientStartEx(&avClientInConfig, &avClientOutConfig);
    if (avIndex < 0) {
        printf("  [%s:%d] avClientStartEx failed avIndex[%d]\n", __FUNCTION__, __LINE__, avIndex);
    	goto __thread_ConnectCCR_exit;
    }
    printf("  [%s:%d] avClientStartEx success avIndex[%d] resend[%d] two_way_streaming[%d] security_mode[%s]\n", __FUNCTION__, __LINE__,
        avIndex, avClientOutConfig.resend, avClientOutConfig.two_way_streaming, (avClientOutConfig.security_mode==AV_SECURITY_SIMPLE)?"AV_SECURITY_SIMPLE":"AV_SECURITY_DTLS");

    if (gDemoOptions.demo_request_ability ){
        printServerAbility(avIndex);
    }

    if (gDemoOptions.demo_change_password ){
        ret = DemoChangePassword(avIndex);
        if(gDemoOptions.demo_change_password || ret != AV_ER_NoERROR)
            goto __thread_ConnectCCR_exit;
    }

    if(gDemoOptions.demo_create_new_token ) {
        ret = DemoCreateNewToken(avIndex);
        if(gDemoOptions.demo_create_new_token || ret != AV_ER_NoERROR)
            goto __thread_ConnectCCR_exit;
    }

    if(gDemoOptions.demo_delete_idenity) {
        DemoDeleteIdentity(avIndex);
        goto __thread_ConnectCCR_exit;
    }

    if(gDemoOptions.demo_list_identity){
        DemoListIdentity(avIndex);
        goto __thread_ConnectCCR_exit;
    }

    ServerInfo->SID = SID;
    ServerInfo->avIndex = avIndex;
    ServerInfo->twoWayStream = avClientOutConfig.two_way_streaming;
#if ENABLE_SPEAKER
    ServerInfo->speakerCh = IOTC_Session_Get_Free_Channel(SID);
#else
    ServerInfo->speakerCh = -1;
#endif

    // Create Video/Audio Recv thread
    ret = pthread_create(&ThreadAVData_ID, NULL, &thread_ReceiveAVData, (void *)&avIndex);
    if (ret != 0) {
        printf("  [%s:%d] pthread_create thread_ReceiveAVData failed\n", __FUNCTION__, __LINE__);
        goto __thread_ConnectCCR_exit;
    }
#if ENABLE_SPEAKER
    // Create Speaker thread
    ret = pthread_create(&ThreadSpeaker_ID, NULL, &thread_Speaker, (void *)ServerInfo);
    if (ret != 0) {
        printf("  [%s:%d] pthread_create thread_Speaker failed\n", __FUNCTION__, __LINE__);
        goto __thread_ConnectCCR_exit;
    }
#endif
    printf("  [%s:%d] start_ipcam_stream\n", __FUNCTION__, __LINE__);
    if(start_ipcam_stream(avIndex, ServerInfo) < 0){
        goto __thread_ConnectCCR_exit;
    }
    if( ThreadAVData_ID!=0){
        pthread_join(ThreadAVData_ID, NULL);
        ThreadAVData_ID = 0;
    }
    if( ThreadSpeaker_ID!=0){
        pthread_join(ThreadSpeaker_ID, NULL);
        ThreadSpeaker_ID = 0;
    }

__thread_ConnectCCR_exit:
	printf("  [%s:%d] ret[%d]\n", __FUNCTION__, __LINE__, ret);
	if(avIndex >= 0){
		avClientStop(avIndex);
		printf("  [%s:%d] avClientStop avIndex[%d]\n", __FUNCTION__, __LINE__, avIndex);
	}
    if(SID >= 0){
		IOTC_Session_Close(SID);
		printf("  [%s:%d] IOTC_Session_Close SID[%d]\n", __FUNCTION__, __LINE__, SID);
	}
    free(arg);

    if( ThreadAVData_ID!=0)
        pthread_join(ThreadAVData_ID, NULL);
    if( ThreadSpeaker_ID!=0)
        pthread_join(ThreadSpeaker_ID, NULL);
    return NULL;
}



int main(int argc, char *argv[])
{
    int ret = 0;

    parsionOptions(argc, argv, &gDemoOptions);

    if (loadConfig(&gDemoConfig) < 0) {
        printf("loadConfig failed exit...!!\n");
        return -1;
    }

    printf("%s start\n", argv[0]);

    printf("IOTCAPI version[%s] AVAPI version[%s]\n", IOTC_Get_Version_String(), avGetAVApiVersionString());
	

	//set the output path of the log file when using log library.
	LogAttr log_attr;
    memset(&log_attr, 0, sizeof(log_attr));

    log_attr.path = "./log_iotc_client.txt";
    ret = IOTC_Set_Log_Attr(log_attr);
    printf("IOTC_Set_Log_Attr ret[%d]\n", ret);

    log_attr.path = "./log_global_client.txt";
    ret = TUTK_Set_Log_Attr(log_attr);
    printf("TUTK_Set_Log_Attr ret[%d]\n", ret);

    log_attr.path = "./log_av_client.txt";
    ret = AV_Set_Log_Attr(log_attr);
    printf("AV_Set_Log_Attr ret[%d]\n", ret);

    const char *license_key = getenv(ENV_LICENSE_KEY);
    if (license_key == NULL) {
        printf("Environment variable %s must be set!!\n", ENV_LICENSE_KEY);
        return -1;
    }
    ret = TUTK_SDK_Set_License_Key(license_key);
    if (ret != TUTK_ER_NoERROR)
    {
        printf("TUTK_SDK_Set_License_Key()=%d ,exit...!!\n", ret);
        return 0;
    }

    srand(time(NULL));

    gettimeofday(&gConnectStartTv, NULL);
    ret = IOTC_Initialize2(0);
    if (ret != IOTC_ER_NoERROR) {
        printf("[%s:%d] IOTC_Initialize2 failed ret[%d]\n", __FUNCTION__, __LINE__, ret);
        PrintErrHandling (ret);
        return -1;
    }

    if (gDemoOptions.demo_lan_search_devices) {
        DemoLanSearch();
        IOTC_DeInitialize();
        return 0;
    }

    // alloc 3 sessions for video and two-way audio
    avInitialize(32);

    pthread_t ConnectThread_ID;
    AV_Server *ServerInfo = (AV_Server *)malloc(sizeof(AV_Server));
    if (ServerInfo == NULL) {
        printf("[%s:%d] malloc ServerInfo failed\n", __FUNCTION__, __LINE__);
        return -1;
    }

    memset(ServerInfo, 0, sizeof(AV_Server));
    strncpy(ServerInfo->UID, gDemoOptions.uid, 21);
    ret = pthread_create(&ConnectThread_ID, NULL, &thread_ConnectCCR, (void *)ServerInfo);
    if (ret != 0) {
        printf("[%s:%d] pthread_create thread_ConnectCCR failed ret[%d]\n", __FUNCTION__, __LINE__, ret);
        free(ServerInfo);
        avDeInitialize();
        IOTC_DeInitialize();
        return -1;
    }
    pthread_join(ConnectThread_ID, NULL);

    avDeInitialize();
    IOTC_DeInitialize();

    printf("%s exit\n", argv[0]);

    return 0;
}


